/********************************************************************
* Copyright Richard Jones June 2002
This file is part of the rspell library.

    The rspell library is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*********************************************************************/
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <malloc.h>
#include "utilities.h"
#include "hash.h"
#include "links.h"
#include "rspell.h"

/*
* Global variables used by this file
*/
static LINK_NODE **hash_table = NULL;	/*This is an array of
										  pointers to LINK_NODEs*/
static char *words = NULL;				/*array of words in the dictionary*/
static char *user_words = NULL;			/*User's word list*/
static LINK_NODE *extra_words = NULL;	/*List for new words*/

/*Predefinitions for functions*/
static void add_to_hash_table(char*);
static void string_to_lower(char *);
static void add_words_to_dictionary(char *);
static int add_new_word_to_hash_table(char *);
static int append_to_file(char *, char *);

#ifndef NDEBUG
static void show_statistics();
static size_t word_count=0;		/*used to check the number of words*/
static size_t extra_words_count=0;
#endif

/******************************************************************
* Initialise the dictionary in memory
*******************************************************************/
int init_dictionary(char *wordlist_file)
{
	/*Check to make sure dictionary is not already initialised*/
	if (hash_table != NULL)
	{
		show_error("Dictionary is already initialised",ZERO);
		return FAIL;
	}
	
	/*First create the hash table*/
	if ((hash_table = (LINK_NODE **) malloc (TABLE_SIZE *sizeof(LINK_NODE *)))
		== NULL)
	{
		show_error("Unable to malloc hash table",errno);
		return FAIL;
	}
	
	/*Fill it with NULL in case malloc doesn't*/
	memset (hash_table, ZERO, TABLE_SIZE * sizeof(LINK_NODE *));

	/*Now get the wordlist from the file and work on it*/
	if (fill_array(wordlist_file,&words))
	{
		add_words_to_dictionary(words);
	}
	else
	{
		/*Failed to get word list into memory*/
		remove_dictionary();
		return FAIL;
	}

	return OK;
}

/******************************************************************
* Add the words from the given wordlist to the dictionary 
*******************************************************************/
int add_user_dictionary(char *user_word_list_file)
{
	/*Check the dictionary has been initialised first*/
	if (hash_table != NULL)
	{
		/*Check we haven't already added a user list*/
		if (user_words != NULL)
		{
			show_error("Cannot add more than one user list",ZERO);
			return FAIL;
		}
		else
		{
			/*Get the wordlist from the file and work on it*/
			if (fill_array(user_word_list_file,&user_words))
			{
				add_words_to_dictionary(user_words);
			}
			else
			{
				/*Failed to get word list into memory*/
				return FAIL;
			}
		}
	}
	else
	{
		show_error("Dictionary has not been initialised",ZERO);
		return FAIL;
	}
	
	return OK;
}

/******************************************************************
* Put the word into the hash table list
*******************************************************************/
static void add_to_hash_table(char *the_word)
{
HASH the_hash;
/* 
* The hash_table points to the first node in the list, or NULL,
* so we can add it directly to itself, as
* the add_node function assumes that NULL means this is the
* first in the list
*/
	the_hash=hash(the_word);
	
	hash_table[the_hash] = add_node(hash_table[the_hash], TO_VOIDP(the_word));

#ifndef NDEBUG
			word_count++;
#endif
	
	assert (hash_table[the_hash] != NULL );
}

/******************************************************************
* Check the string given to see if it is in the dictionary
* return 0 if it is not or there is a failure, 1 if it is
*******************************************************************/
int rspell_check(char *word)
{
char *the_word;
LINK_NODE *the_node;
int found = 0;
int length;
	
	/*Duplicate the word (strdup does not exist on all system,
	so do it explicitly)*/
	length=strlen (word);
	the_word=malloc((length+1)*sizeof(char));
		
	if (the_word != NULL)
	{
		/*Copy the word, along with the null character*/
		strncpy(the_word,word,length+1);
		
		/*set the duplicated word to lower case*/
		string_to_lower(the_word);

		/*
		* Look at all the words at the hash value to see if it is correct
		* Check hash_table has been initialised first!
		* If it hasn't, say the word is not in the list (ie don't crash!)
		*/
		if (hash_table != NULL)
		{
			the_node = hash_table[hash(the_word)];
		}
		else
		{
			the_node = NULL;
		}

		while ((the_node != NULL) && !found)
		{
			if (strncmp(TO_CHARP(the_node->word), the_word, strlen(the_word)) == 0)
			{
				found = 1;
			}
			the_node = the_node->next;
		}
		
		/*free the memory used by the duplicate word*/
		free (the_word);
	}
	
	return found;
}

/******************************************************************
* Convert a string to lower case value chars only
*******************************************************************/
static void string_to_lower(char *sptr)
{
/*
* Note that this function actually changes the string passed to it
*/
	while ( *sptr != '\0')
	{
		if ( (*sptr >= 'A') && (*sptr <= 'Z'))
		{
			*sptr |= 32;
		}
		sptr++;
	}
}

/******************************************************************
* Free all memory used by the dictionary
* No need to return a value, as we make sure it can't go wrong
* We always set pointers to NULL after freeing them
* so the software can re-initialise things without failing
*******************************************************************/
void remove_dictionary()
{
long i;
#ifndef NDEBUG
size_t count=0;
size_t e_count=0;
#endif

#ifndef NDEBUG
	show_statistics();
#endif

	/*We can get rid of the wordlists first*/
	if (words != NULL)
	{
		free (words);
		words = NULL;
	}
	if (user_words != NULL)
	{
		free (user_words);
		user_words = NULL;
	}
	
	/*
	* Tidy up the extra words
	*/
	if (extra_words != NULL)
	{
		/*Free the memory taken up by its data pointers*/
		remove_words_from_list(extra_words);
		
		/*Now delete the list itself*/
#ifndef NDEBUG
		e_count=
#endif
		remove_list(extra_words);
		extra_words = NULL;	

		assert(e_count == extra_words_count);
		
		extra_words_count = 0;
	}
	
	/*
	* Free each of the list nodes by running through the table
	*/
	if (hash_table != NULL)
	{
		for (i=0; i<TABLE_SIZE; i++)
		{
#ifndef NDEBUG
			count +=
#endif
			remove_list(hash_table[i]);
			hash_table[i]=NULL;

		}

		/*
		* Check (on debug) we removed the same number
		* of nodes as we put in in the first place
		*/
		assert (count == word_count);

#ifndef NDEBUG
			/*Reset so will still work if another dictionary is loaded*/
			word_count =0;
#endif
		/*Finally remove the hash table*/
		free (hash_table);
		hash_table = NULL;
	}

}

/******************************************************************
* Take a list of words seperated by \n.
* Parse through, adding each word to the hash table
* Replace the \n with \0 so can use the same memory without
* needing to copy the word.
*******************************************************************/
void add_words_to_dictionary(char *wordlist)
{
char *the_word; /*word that will be added to hash table*/
char *token;		/*used when parsing the word array*/

		/*split array into words for use*/
		token=strchr(wordlist,'\n');

		the_word=wordlist;
		
		do
		{
			*token++='\0';	/*
							* make the \n an "end of word" instead
							* so the previous pointer can be used as a 
							* null terminated string
							*/
			/*Use the word here*/
			add_to_hash_table(the_word);
			
			the_word=token;

		}while((token=strchr(token,'\n')) != NULL);
}

#ifndef NDEBUG
/******************************************************************
* Print out some statistics on the dictionary
* only compiles stuff if debug is on
*******************************************************************/
void show_statistics()
{
/*Used for debug output at end of program*/
unsigned long counter;
long list_length;
long max_num_nodes=0;
long empty_hashes=0;

	/*
	* Make sure that the hash table has been initialised!
	*/
	if (hash_table != NULL)
	{
		for (counter=0;counter<TABLE_SIZE;counter++)
		{
			list_length=nodes_in_list(hash_table[counter]);
			if (list_length > max_num_nodes)
				max_num_nodes = list_length;
			if (0 == list_length)
				empty_hashes++;
		}
		printf("Collisions detected on %d words = %ld\n",word_count,num_collisions);
		printf("Maximum number of words at any hash value= %ld\n",max_num_nodes);
		printf("Number of empty hashes = %ld\n",empty_hashes);
	}
	else
	{
		show_error("No Stats available on uninitialised dictionary",ZERO);
	}
}
#endif

/************************************************************************
* Add a new word to the extra word list and put it into the hash table
*************************************************************************/
int add_new_word_to_hash_table(char *new_word)
{
char *word_ptr;
size_t length;

	/*First check the hash table is OK*/
	if (hash_table != NULL)
	{
		length = strlen(new_word);
		/*Allocate memory for the copy of the word*/
		word_ptr = (char *) malloc((length+1)*sizeof(char));
		if (word_ptr != NULL)
		{
			/*copy in the new word, with \0 at end*/
			strncpy(word_ptr, new_word, length+1);

			/*Set it to lower case*/
			string_to_lower(word_ptr);

			/*add it to the extra word list*/
			extra_words = add_node(extra_words, TO_VOIDP(word_ptr));
			
#ifndef NDEBUG
			extra_words_count++;
#endif

			/*add it to the hash table*/
			add_to_hash_table(word_ptr);
		}
		else
		{
			show_error("Unable to create space for new word",errno);
			return FAIL;
		}
	}
	else
	{
		show_error("Dictionary not properly initialised",ZERO);
		return FAIL;
	}

	return OK;
}

/***************************************************************************
* Add word to dictionary
* This is temporary and is orgotten when dictionary is removed
* Called by the user...
****************************************************************************/
int add_new_word_temp(char *new_word)
{
	if (new_word != NULL)
	{
	add_new_word_to_hash_table(new_word);
	}
	else
	{
		show_error("Invalid word given",ZERO);
		return FAIL;
	}
	return OK;
}

/***************************************************************************
* Append the word to the filename
* Create file if it doesn't exist
****************************************************************************/
int append_to_file(char *filename, char *new_word)
{
FILE *out;
int out_error;

	if ((NULL == filename) || (NULL == new_word))
	{
		show_error("Invalid filename of new word given", ZERO);
		return FAIL;
	}
	else
	{
		/*
		* Try to open the file to append to it
		* Use binary mode for better portability, so we can
		* output UTF-8 stuff as well as ASCII-some systems 
		* distinguish between bytes and chars
		*/
		out=fopen(filename,"ab");
		
		if (NULL == out)
		{
			show_error("Unable to open user word list",errno);
			return FAIL;
		}
		else
		{
			out_error=OK;
			
			/*Write the word followed by end ofline*/
			if (fputs(new_word,out) < 0)
			{
				out_error = FAIL;
			}
			if (fputs("\n",out) < 0)
			{
				out_error = FAIL;
			}
			/*Close the file before continuing*/
			fclose(out);
			
			/*Check we did ok*/
			if (FAIL == out_error)
			{
				show_error("Unable to save new word",errno);
				return FAIL;
			}
		}
	}
	
	return OK;
}

/***************************************************************************
* Add new word to user list
* Create file if it doesn't exist
* Calld by user...
****************************************************************************/
int add_new_word_perm(char *filename, char *new_word)
{
size_t length;
char *word_ptr;

	length = strlen(new_word);
	/*Allocate memory for the copy of the word*/
	word_ptr = (char *) malloc((length+1)*sizeof(char));
	if (word_ptr != NULL)
	{
		/*copy in the new word, with \0 at end*/
		strncpy(word_ptr, new_word, length+1);

		/*Set it to lower case*/
		string_to_lower(word_ptr);

		/*Append it to the file*/
		if (!append_to_file(filename,word_ptr))
		{
			show_error("Could not append to file",ZERO);
			return FAIL;
		}
	}
	else
	{
		show_error("Unable to get memory to copy word",errno);
		return FAIL;
	}
		
	return OK;
}
